@@ -25,7 +25,6 @@ from .helpers import BasicAuth
 
 __all__ = ('BaseConnector', 'TCPConnector', 'ProxyConnector', 'UnixConnector')
 
-PY_341 = sys.version_info >= (3, 4, 1)
 PY_343 = sys.version_info >= (3, 4, 3)
 
 HASHFUNC_BY_DIGESTLEN = {
@@ -53,24 +52,23 @@ class Connection(object):
         if loop.get_debug():
             self._source_traceback = traceback.extract_stack(sys._getframe(1))
 
-    if PY_341:
-        def __del__(self):
-            if self._transport is not None:
-                if hasattr(self._loop, 'is_closed'):
-                    if self._loop.is_closed():
-                        return
+    def __del__(self, _warnings=warnings):
+        if self._transport is not None:
+            _warnings.warn("Unclosed connection {!r}".format(self),
+                           ResourceWarning)
+            if hasattr(self._loop, 'is_closed'):
+                if self._loop.is_closed():
+                    return
 
-                self._connector._release(
-                    self._key, self._request, self._transport, self._protocol,
-                    should_close=True)
+            self._connector._release(
+                self._key, self._request, self._transport, self._protocol,
+                should_close=True)
 
-                warnings.warn("Unclosed connection {!r}".format(self),
-                              ResourceWarning)
-                context = {'client_connection': self,
-                           'message': 'Unclosed connection'}
-                if self._source_traceback is not None:
-                    context['source_traceback'] = self._source_traceback
-                self._loop.call_exception_handler(context)
+            context = {'client_connection': self,
+                       'message': 'Unclosed connection'}
+            if self._source_traceback is not None:
+                context['source_traceback'] = self._source_traceback
+            self._loop.call_exception_handler(context)
 
     @property
     def loop(self):
@@ -142,22 +140,21 @@ class BaseConnector(object):
 
         self.cookies = http.cookies.SimpleCookie()
 
-    if PY_341:
-        def __del__(self):
-            if self._closed:
-                return
-            if not self._conns:
-                return
+    def __del__(self, _warnings=warnings):
+        if self._closed:
+            return
+        if not self._conns:
+            return
 
-            self.close()
+        self.close()
 
-            warnings.warn("Unclosed connector {!r}".format(self),
-                          ResourceWarning)
-            context = {'connector': self,
-                       'message': 'Unclosed connector'}
-            if self._source_traceback is not None:
-                context['source_traceback'] = self._source_traceback
-            self._loop.call_exception_handler(context)
+        _warnings.warn("Unclosed connector {!r}".format(self),
+                       ResourceWarning)
+        context = {'connector': self,
+                   'message': 'Unclosed connector'}
+        if self._source_traceback is not None:
+            context['source_traceback'] = self._source_traceback
+        self._loop.call_exception_handler(context)
 
     @property
     def force_close(self):
@@ -293,10 +290,13 @@ class BaseConnector(object):
 
             except asyncio.TimeoutError as exc:
                 raise ClientTimeoutError(
-                    'Connection timeout to host %s:%s ssl:%s' % key) from exc
+                    'Connection timeout to host {0[0]}:{0[1]} ssl:{0[2]}'
+                    .format(key)) from exc
             except OSError as exc:
                 raise ClientOSError(
-                    'Cannot connect to host %s:%s ssl:%s' % key) from exc
+                    exc.errno,
+                    'Cannot connect to host {0[0]}:{0[1]} ssl:{0[2]} [{1}]'
+                    .format(key, exc.strerror)) from exc
 
         self._acquired[key].add(transport)
         conn = Connection(self, key, req, transport, proto, self._loop)
@@ -347,14 +347,10 @@ class BaseConnector(object):
         resp = req.response
 
         if not should_close:
-            if resp is not None:
-                if resp.message is None:
-                    should_close = True
-                else:
-                    should_close = resp.message.should_close
-
             if self._force_close:
                 should_close = True
+            elif resp is not None:
+                should_close = resp._should_close
 
         reader = protocol.reader
         if should_close or (reader.output and not reader.output.at_eof()):
@@ -373,9 +369,6 @@ class BaseConnector(object):
         raise NotImplementedError()
 
 
-_SSL_OP_NO_COMPRESSION = getattr(ssl, "OP_NO_COMPRESSION", 0)
-_SSH_HAS_CREATE_DEFAULT_CONTEXT = hasattr(ssl, 'create_default_context')
-
 _marker = object()
 
 
@@ -395,7 +388,7 @@ class TCPConnector(BaseConnector):
 
     def __init__(self, *, verify_ssl=True, fingerprint=None,
                  resolve=_marker, use_dns_cache=_marker,
-                 family=socket.AF_INET, ssl_context=None,
+                 family=0, ssl_context=None,
                  **kwargs):
         super().__init__(**kwargs)
 
@@ -456,19 +449,10 @@ class TCPConnector(BaseConnector):
                 sslcontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                 sslcontext.options |= ssl.OP_NO_SSLv2
                 sslcontext.options |= ssl.OP_NO_SSLv3
-                sslcontext.options |= _SSL_OP_NO_COMPRESSION
+                sslcontext.options |= ssl.OP_NO_COMPRESSION
                 sslcontext.set_default_verify_paths()
-            elif _SSH_HAS_CREATE_DEFAULT_CONTEXT:
-                # Python 3.4+
-                sslcontext = ssl.create_default_context()
             else:
-                # Fallback for Python 3.3.
-                sslcontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
-                sslcontext.options |= ssl.OP_NO_SSLv2
-                sslcontext.options |= ssl.OP_NO_SSLv3
-                sslcontext.options |= _SSL_OP_NO_COMPRESSION
-                sslcontext.set_default_verify_paths()
-                sslcontext.verify_mode = ssl.CERT_REQUIRED
+                sslcontext = ssl.create_default_context()
             self._ssl_context = sslcontext
         return self._ssl_context
 
@@ -572,6 +556,11 @@ class TCPConnector(BaseConnector):
                 has_cert = transp.get_extra_info('sslcontext')
                 if has_cert and self._fingerprint:
                     sock = transp.get_extra_info('socket')
+                    if not hasattr(sock, 'getpeercert'):
+                        # Workaround for asyncio 3.5.0
+                        # Starting from 3.5.1 version
+                        # there is 'ssl_object' extra info in transport
+                        sock = transp._ssl_protocol._sslpipe.ssl_object
                     # gives DER-encoded cert as a sequence of bytes (or None)
                     cert = sock.getpeercert(binary_form=True)
                     assert cert
@@ -584,8 +573,9 @@ class TCPConnector(BaseConnector):
             except OSError as e:
                 exc = e
         else:
-            raise ClientOSError('Can not connect to %s:%s' %
-                                (req.host, req.port)) from exc
+            raise ClientOSError(exc.errno,
+                                'Can not connect to %s:%s [%s]' %
+                                (req.host, req.port, exc.strerror)) from exc
 
 
 class ProxyConnector(TCPConnector):
