@@ -2,11 +2,12 @@
 import base64
 import io
 import os
-import urllib.parse
+from urllib.parse import quote, urlencode
 from collections import namedtuple
 from wsgiref.handlers import format_date_time
 
 from . import hdrs, multidict
+from .errors import InvalidURL
 
 __all__ = ('BasicAuth', 'FormData', 'parse_mimetype')
 
@@ -126,7 +127,7 @@ class FormData:
         for type_options, _, value in self._fields:
             data.append((type_options['name'], value))
 
-        data = urllib.parse.urlencode(data, doseq=True)
+        data = urlencode(data, doseq=True)
         return data.encode(encoding)
 
     def _gen_form_data(self, *args, **kwargs):
@@ -238,7 +239,7 @@ def atoms(message, environ, response, transport, request_time):
 
     if transport is not None:
         remote_addr = parse_remote_addr(
-            transport.get_extra_info('addr', '127.0.0.1'))
+            transport.get_extra_info('peername', ('127.0.0.1', )))
     else:
         remote_addr = ('',)
 
@@ -253,7 +254,7 @@ def atoms(message, environ, response, transport, request_time):
         'f': headers.get(hdrs.REFERER, '-'),
         'a': headers.get(hdrs.USER_AGENT, '-'),
         'T': str(int(request_time)),
-        'D': str(request_time).split('.', 1)[-1][:5],
+        'D': str(request_time).split('.', 1)[-1][:6],
         'p': "<%s>" % os.getpid()
     }
 
@@ -290,6 +291,9 @@ class SafeAtoms(dict):
             return '-'
 
 
+_marker = object()
+
+
 class reify:
     """Use as a class method decorator.  It operates almost exactly like
     the Python ``@property`` decorator, but it puts the result of the
@@ -305,10 +309,65 @@ class reify:
             self.__doc__ = wrapped.__doc__
         except:  # pragma: no cover
             pass
+        self.name = wrapped.__name__
 
-    def __get__(self, inst, objtype=None):
+    def __get__(self, inst, owner, _marker=_marker):
         if inst is None:
             return self
+        val = inst.__dict__.get(self.name, _marker)
+        if val is not _marker:
+            return val
         val = self.wrapped(inst)
-        setattr(inst, self.wrapped.__name__, val)
+        inst.__dict__[self.name] = val
         return val
+
+    def __set__(self, inst, value):
+        raise AttributeError("reified property is read-only")
+
+
+# The unreserved URI characters (RFC 3986)
+UNRESERVED_SET = frozenset(
+    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz" +
+    "0123456789-._~")
+
+
+def unquote_unreserved(uri):
+    """Un-escape any percent-escape sequences in a URI that are unreserved
+    characters. This leaves all reserved, illegal and non-ASCII bytes encoded.
+    """
+    parts = uri.split('%')
+    for i in range(1, len(parts)):
+        h = parts[i][0:2]
+        if len(h) == 2 and h.isalnum():
+            try:
+                c = chr(int(h, 16))
+            except ValueError:
+                raise InvalidURL("Invalid percent-escape sequence: '%s'" % h)
+
+            if c in UNRESERVED_SET:
+                parts[i] = c + parts[i][2:]
+            else:
+                parts[i] = '%' + parts[i]
+        else:
+            parts[i] = '%' + parts[i]
+    return ''.join(parts)
+
+
+def requote_uri(uri):
+    """Re-quote the given URI.
+
+    This function passes the given URI through an unquote/quote cycle to
+    ensure that it is fully and consistently quoted.
+    """
+    safe_with_percent = "!#$%&'()*+,/:;=?@[]~"
+    safe_without_percent = "!#$&'()*+,/:;=?@[]~"
+    try:
+        # Unquote only the unreserved characters
+        # Then quote only illegal characters (do not quote reserved,
+        # unreserved, or '%')
+        return quote(unquote_unreserved(uri), safe=safe_with_percent)
+    except InvalidURL:
+        # We couldn't unquote the given URI, so let's try quoting it, but
+        # there may be unquoted '%'s in the URI. We need to make sure they're
+        # properly quoted so they do not cause issues elsewhere.
+        return quote(uri, safe=safe_without_percent)
