@@ -168,7 +168,7 @@ class ServerHttpProtocol(aiohttp.StreamProtocol):
             self._timeout_handle = None
 
     def data_received(self, data):
-        self.reader.feed_data(data)
+        super().data_received(data)
 
         # reading request
         if not self._reading_request:
@@ -271,27 +271,32 @@ class ServerHttpProtocol(aiohttp.StreamProtocol):
 
                 yield from self.handle_request(message, payload)
 
-            except (asyncio.CancelledError, errors.ClientDisconnectedError):
-                if self.debug:
-                    self.log_exception(
-                        'Ignored premature client disconnection.')
-                break
+            except asyncio.CancelledError:
+                return
+            except errors.ClientDisconnectedError:
+                self.log_debug(
+                    'Ignored premature client disconnection #1.')
+                return
             except errors.HttpProcessingError as exc:
                 if self.transport is not None:
                     yield from self.handle_error(exc.code, message,
-                                                 None, exc, exc.headers)
+                                                 None, exc, exc.headers,
+                                                 exc.message)
+            except errors.LineLimitExceededParserError as exc:
+                yield from self.handle_error(400, message, None, exc)
             except Exception as exc:
                 yield from self.handle_error(500, message, None, exc)
             finally:
                 if self.transport is None:
-                    self.log_debug('Ignored premature client disconnection.')
-                    break
+                    self.log_debug(
+                        'Ignored premature client disconnection #2.')
+                    return
 
                 if payload and not payload.is_eof():
                     self.log_debug('Uncompleted request.')
                     self._request_handler = None
                     self.transport.close()
-                    break
+                    return
                 else:
                     reader.unset_parser()
 
@@ -309,13 +314,13 @@ class ServerHttpProtocol(aiohttp.StreamProtocol):
                         self.log_debug('Close client connection.')
                         self._request_handler = None
                         self.transport.close()
-                        break
+                        return
                 else:
                     # connection is closed
-                    break
+                    return
 
-    def handle_error(self, status=500,
-                     message=None, payload=None, exc=None, headers=None):
+    def handle_error(self, status=500, message=None,
+                     payload=None, exc=None, headers=None, reason=None):
         """Handle errors.
 
         Returns http response with specific status code. Logs additional
@@ -330,7 +335,10 @@ class ServerHttpProtocol(aiohttp.StreamProtocol):
                 self.log_exception("Error handling request")
 
             try:
-                reason, msg = RESPONSES[status]
+                if reason is None or reason == '':
+                    reason, msg = RESPONSES[status]
+                else:
+                    msg = reason
             except KeyError:
                 status = 500
                 reason, msg = '???', ''
