@@ -15,7 +15,7 @@ def intro(request):
     binary = txt.encode('utf8')
     resp = StreamResponse()
     resp.content_length = len(binary)
-    resp.start(request)
+    yield from resp.prepare(request)
     resp.write(binary)
     return resp
 
@@ -36,7 +36,7 @@ def hello(request):
     name = request.match_info.get('name', 'Anonymous')
     answer = ('Hello, ' + name).encode('utf8')
     resp.content_length = len(answer)
-    resp.start(request)
+    yield from resp.prepare(request)
     resp.write(answer)
     yield from resp.write_eof()
     return resp
@@ -51,10 +51,14 @@ def init(loop):
     app.router.add_route('GET', '/hello/{name}', hello)
     app.router.add_route('GET', '/hello', hello)
 
-    srv = yield from loop.create_server(app.make_handler(), '127.0.0.1', 8080)
+    handler = app.make_handler()
+    srv = yield from loop.create_server(handler, '127.0.0.1', 8080)
     print("Server started at http://127.0.0.1:8080")
-    return srv
+    return srv, handler
 
 loop = asyncio.get_event_loop()
-loop.run_until_complete(init(loop))
-loop.run_forever()
+srv, handler = loop.run_until_complete(init(loop))
+try:
+    loop.run_forever()
+except KeyboardInterrupt:
+    loop.run_until_complete(handler.finish_connections())
