@@ -6,104 +6,17 @@ import gc
 import socket
 import unittest
 import ssl
-import sys
+import tempfile
+import shutil
+import os.path
 from unittest import mock
 
 import aiohttp
+from aiohttp import web
 from aiohttp import client
-from aiohttp import test_utils
-from aiohttp.errors import FingerprintMismatch
 from aiohttp.client import ClientResponse, ClientRequest
 from aiohttp.connector import Connection
 
-from tests.test_client_functional import Functional
-
-PY_341 = sys.version_info >= (3, 4, 1)
-
-
-class TestHttpConnection(unittest.TestCase):
-
-    def setUp(self):
-        self.key = object()
-        self.connector = mock.Mock()
-        self.request = mock.Mock()
-        self.transport = mock.Mock()
-        self.protocol = mock.Mock()
-        self.loop = asyncio.new_event_loop()
-        asyncio.set_event_loop(None)
-
-    def tearDown(self):
-        self.loop.close()
-
-    @unittest.skipUnless(PY_341, "Requires Python 3.4.1+")
-    def test_del(self):
-        conn = Connection(
-            self.connector, self.key, self.request,
-            self.transport, self.protocol, self.loop)
-        exc_handler = unittest.mock.Mock()
-        self.loop.set_exception_handler(exc_handler)
-
-        with self.assertWarns(ResourceWarning):
-            del conn
-            gc.collect()
-
-        self.connector._release.assert_called_with(self.key,
-                                                   self.request,
-                                                   self.transport,
-                                                   self.protocol,
-                                                   should_close=True)
-        msg = {'client_connection': unittest.mock.ANY,  # conn was deleted
-               'message': 'Unclosed connection'}
-        if self.loop.get_debug():
-            msg['source_traceback'] = unittest.mock.ANY
-        exc_handler.assert_called_with(self.loop, msg)
-
-    def test_close(self):
-        conn = Connection(
-            self.connector, self.key, self.request,
-            self.transport, self.protocol, self.loop)
-        self.assertFalse(conn.closed)
-        conn.close()
-        self.assertIsNone(conn._transport)
-        self.connector._release.assert_called_with(
-            self.key, self.request, self.transport, self.protocol,
-            should_close=True)
-        self.assertTrue(conn.closed)
-
-    def test_release(self):
-        conn = Connection(
-            self.connector, self.key, self.request,
-            self.transport, self.protocol, self.loop)
-        self.assertFalse(conn.closed)
-        conn.release()
-        self.assertFalse(self.transport.close.called)
-        self.assertIsNone(conn._transport)
-        self.connector._release.assert_called_with(
-            self.key, self.request, self.transport, self.protocol,
-            should_close=False)
-        self.assertTrue(conn.closed)
-
-    def test_release_released(self):
-        conn = Connection(
-            self.connector, self.key, self.request,
-            self.transport, self.protocol, self.loop)
-        conn.release()
-        self.connector._release.reset_mock()
-        conn.release()
-        self.assertFalse(self.transport.close.called)
-        self.assertIsNone(conn._transport)
-        self.assertFalse(self.connector._release.called)
-
-    def test_detach(self):
-        conn = Connection(
-            self.connector, self.key, self.request,
-            self.transport, self.protocol, self.loop)
-        self.assertFalse(conn.closed)
-        conn.detach()
-        self.assertIsNone(conn._transport)
-        self.assertFalse(self.connector._release.called)
-        self.assertTrue(conn.closed)
-
 
 class TestBaseConnector(unittest.TestCase):
 
@@ -113,14 +26,14 @@ class TestBaseConnector(unittest.TestCase):
 
         self.transport = unittest.mock.Mock()
         self.stream = aiohttp.StreamParser()
-        self.response = ClientResponse('get', 'http://python.org')
-        self.response._loop = self.loop
+        self.response = ClientResponse('get', 'http://base-conn.org')
+        self.response._post_init(self.loop)
 
     def tearDown(self):
+        self.response.close()
         self.loop.close()
         gc.collect()
 
-    @unittest.skipUnless(PY_341, "Requires Python 3.4.1+")
     def test_del(self):
         conn = aiohttp.BaseConnector(loop=self.loop)
         transp = unittest.mock.Mock()
@@ -142,7 +55,6 @@ class TestBaseConnector(unittest.TestCase):
             msg['source_traceback'] = unittest.mock.ANY
         exc_handler.assert_called_with(self.loop, msg)
 
-    @unittest.skipUnless(PY_341, "Requires Python 3.4.1+")
     def test_del_with_scheduled_cleanup(self):
         conn = aiohttp.BaseConnector(loop=self.loop, keepalive_timeout=0.01)
         transp = unittest.mock.Mock()
@@ -166,7 +78,6 @@ class TestBaseConnector(unittest.TestCase):
             msg['source_traceback'] = unittest.mock.ANY
         exc_handler.assert_called_with(self.loop, msg)
 
-    @unittest.skipUnless(PY_341, "Requires Python 3.4.1+")
     def test_del_with_closed_loop(self):
         conn = aiohttp.BaseConnector(loop=self.loop)
         transp = unittest.mock.Mock()
@@ -186,7 +97,6 @@ class TestBaseConnector(unittest.TestCase):
         self.assertFalse(transp.close.called)
         self.assertTrue(exc_handler.called)
 
-    @unittest.skipUnless(PY_341, "Requires Python 3.4.1+")
     def test_del_empty_conector(self):
         conn = aiohttp.BaseConnector(loop=self.loop)
 
@@ -249,7 +159,7 @@ class TestBaseConnector(unittest.TestCase):
         conn._start_cleanup_task = unittest.mock.Mock()
         req = unittest.mock.Mock()
         resp = req.response = unittest.mock.Mock()
-        resp.message.should_close = False
+        resp._should_close = False
 
         tr, proto = unittest.mock.Mock(), unittest.mock.Mock()
         key = 1
@@ -387,11 +297,15 @@ class TestBaseConnector(unittest.TestCase):
         conn = aiohttp.BaseConnector(loop=self.loop)
         conn._create_connection = unittest.mock.Mock()
         conn._create_connection.return_value = asyncio.Future(loop=self.loop)
-        conn._create_connection.return_value.set_exception(OSError())
+        err = OSError(1, 'permission error')
+        conn._create_connection.return_value.set_exception(err)
 
-        with self.assertRaises(aiohttp.ClientOSError):
+        with self.assertRaises(aiohttp.ClientOSError) as ctx:
             req = unittest.mock.Mock()
             self.loop.run_until_complete(conn.connect(req))
+        self.assertEqual(1, ctx.exception.errno)
+        self.assertTrue(ctx.exception.strerror.startswith('Cannot connect to'))
+        self.assertTrue(ctx.exception.strerror.endswith('[permission error]'))
 
     def test_start_cleanup_task(self):
         loop = unittest.mock.Mock()
@@ -469,7 +383,7 @@ class TestBaseConnector(unittest.TestCase):
             self.assertFalse(conn.resolve)
         self.assertFalse(conn.use_dns_cache)
 
-        self.assertEqual(conn.family, socket.AF_INET)
+        self.assertEqual(conn.family, 0)
 
         with self.assertWarns(DeprecationWarning):
             self.assertEqual(conn.resolved_hosts, {})
@@ -485,45 +399,6 @@ class TestBaseConnector(unittest.TestCase):
         with self.assertRaises(ValueError):
             aiohttp.TCPConnector(loop=self.loop, fingerprint=invalid)
 
-    def test_tcp_connector_fingerprint(self):
-        # The even-index fingerprints below are "expect success" cases
-        # for ./sample.crt.der, the cert presented by test_utils.run_server.
-        # The odd-index fingerprints are "expect fail" cases.
-        testcases = (
-            # md5
-            b'\xa2\x06G\xad\xaa\xf5\xd8\\J\x99^by;\x06=',
-            b'\x00' * 16,
-
-            # sha1
-            b's\x93\xfd:\xed\x08\x1do\xa9\xaeq9\x1a\xe3\xc5\x7f\x89\xe7l\xf9',
-            b'\x00' * 20,
-
-            # sha256
-            b'0\x9a\xc9D\x83\xdc\x91\'\x88\x91\x11\xa1d\x97\xfd\xcb~7U\x14D@L'
-            b'\x11\xab\x99\xa8\xae\xb7\x14\xee\x8b',
-            b'\x00' * 32,
-        )
-        for i, fingerprint in enumerate(testcases):
-            expect_fail = i % 2
-            conn = aiohttp.TCPConnector(loop=self.loop, verify_ssl=False,
-                                        fingerprint=fingerprint)
-            with test_utils.run_server(self.loop, use_ssl=True) as httpd:
-                coro = client.request('get', httpd.url('method', 'get'),
-                                      connector=conn, loop=self.loop)
-                if expect_fail:
-                    with self.assertRaises(FingerprintMismatch) as cm:
-                        self.loop.run_until_complete(coro)
-                    exc = cm.exception
-                    self.assertEqual(exc.expected, fingerprint)
-                    # the previous test case should be what we actually got
-                    self.assertEqual(exc.got, testcases[i-1])
-                else:
-                    # should not raise
-                    resp = self.loop.run_until_complete(coro)
-                    resp.close(force=True)
-
-            conn.close()
-
     def test_tcp_connector_clear_resolved_hosts(self):
         conn = aiohttp.TCPConnector(loop=self.loop)
         info = object()
@@ -742,46 +617,90 @@ class TestBaseConnector(unittest.TestCase):
 class TestHttpClientConnector(unittest.TestCase):
 
     def setUp(self):
+        self.handler = None
         self.loop = asyncio.new_event_loop()
         asyncio.set_event_loop(None)
 
     def tearDown(self):
-        # just in case if we have transport close callbacks
-        test_utils.run_briefly(self.loop)
-
+        if self.handler:
+            self.loop.run_until_complete(self.handler.finish_connections())
+        self.loop.stop()
+        self.loop.run_forever()
         self.loop.close()
         gc.collect()
 
+    def find_unused_port(self):
+        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        s.bind(('127.0.0.1', 0))
+        port = s.getsockname()[1]
+        s.close()
+        return port
+
+    @asyncio.coroutine
+    def create_server(self, method, path, handler):
+        app = web.Application(loop=self.loop)
+        app.router.add_route(method, path, handler)
+
+        port = self.find_unused_port()
+        self.handler = app.make_handler(keep_alive_on=False)
+        srv = yield from self.loop.create_server(
+            self.handler, '127.0.0.1', port)
+        url = "http://127.0.0.1:{}".format(port) + path
+        self.addCleanup(srv.close)
+        return app, srv, url
+
+    @asyncio.coroutine
+    def create_unix_server(self, method, path, handler):
+        tmpdir = tempfile.mkdtemp()
+        self.addCleanup(shutil.rmtree, tmpdir)
+        app = web.Application(loop=self.loop)
+        app.router.add_route(method, path, handler)
+
+        self.handler = app.make_handler(keep_alive_on=False)
+        sock_path = os.path.join(tmpdir, 'socket.sock')
+        srv = yield from self.loop.create_unix_server(
+            self.handler, sock_path)
+        url = "http://127.0.0.1" + path
+        self.addCleanup(srv.close)
+        return app, srv, url, sock_path
+
     def test_tcp_connector(self):
-        with test_utils.run_server(self.loop, router=Functional) as httpd:
-            r = self.loop.run_until_complete(
-                client.request(
-                    'get', httpd.url('method', 'get'),
-                    connector=aiohttp.TCPConnector(loop=self.loop),
-                    loop=self.loop))
-            content = self.loop.run_until_complete(r.content.read())
-            content = content.decode()
-            self.assertEqual(r.status, 200)
-            r.close()
+        @asyncio.coroutine
+        def handler(request):
+            return web.HTTPOk()
+
+        app, srv, url = self.loop.run_until_complete(
+            self.create_server('get', '/', handler))
+        conn = aiohttp.TCPConnector(loop=self.loop)
+        r = self.loop.run_until_complete(
+            aiohttp.request(
+                'get', url,
+                connector=conn,
+                loop=self.loop))
+        self.loop.run_until_complete(r.release())
+        self.assertEqual(r.status, 200)
+        r.close()
+        conn.close()
 
     @unittest.skipUnless(hasattr(socket, 'AF_UNIX'), 'requires unix')
     def test_unix_connector(self):
-        path = '/tmp/aiohttp_unix.sock'
+        @asyncio.coroutine
+        def handler(request):
+            return web.HTTPOk()
 
-        connector = aiohttp.UnixConnector(path, loop=self.loop)
-        self.assertEqual(path, connector.path)
+        app, srv, url, sock_path = self.loop.run_until_complete(
+            self.create_unix_server('get', '/', handler))
 
-        with test_utils.run_server(
-                self.loop, listen_addr=path, router=Functional) as httpd:
-            r = self.loop.run_until_complete(
-                client.request(
-                    'get', httpd.url('method', 'get'),
-                    connector=connector,
-                    loop=self.loop))
-            content = self.loop.run_until_complete(r.content.read())
-            content = content.decode()
-            self.assertEqual(r.status, 200)
-            r.close()
+        connector = aiohttp.UnixConnector(sock_path, loop=self.loop)
+        self.assertEqual(sock_path, connector.path)
+
+        r = self.loop.run_until_complete(
+            client.request(
+                'get', url,
+                connector=connector,
+                loop=self.loop))
+        self.assertEqual(r.status, 200)
+        r.close()
 
     def test_connector_cookie_deprecation(self):
         with self.assertWarnsRegex(DeprecationWarning,
@@ -817,8 +736,8 @@ class TestProxyConnector(unittest.TestCase):
 
     def tearDown(self):
         # just in case if we have transport close callbacks
-        test_utils.run_briefly(self.loop)
-
+        self.loop.stop()
+        self.loop.run_forever()
         self.loop.close()
         gc.collect()
 
@@ -1022,9 +941,9 @@ class TestProxyConnector(unittest.TestCase):
         tr.pause_reading.assert_called_once_with()
         tr.get_extra_info.assert_called_with('socket', default=None)
 
-        proxy_req.close()
+        self.loop.run_until_complete(proxy_req.close())
         proxy_resp.close()
-        req.close()
+        self.loop.run_until_complete(req.close())
 
     @unittest.mock.patch('aiohttp.connector.ClientRequest')
     def test_https_connect_runtime_error(self, ClientRequestMock):
@@ -1052,9 +971,9 @@ class TestProxyConnector(unittest.TestCase):
                 RuntimeError, "Transport does not expose socket instance"):
             self.loop.run_until_complete(connector._create_connection(req))
 
-        proxy_req.close()
+        self.loop.run_until_complete(proxy_req.close())
         proxy_resp.close()
-        req.close()
+        self.loop.run_until_complete(req.close())
 
     @unittest.mock.patch('aiohttp.connector.ClientRequest')
     def test_https_connect_http_proxy_error(self, ClientRequestMock):
@@ -1083,9 +1002,9 @@ class TestProxyConnector(unittest.TestCase):
                 aiohttp.HttpProxyError, "400, message='bad request'"):
             self.loop.run_until_complete(connector._create_connection(req))
 
-        proxy_req.close()
+        self.loop.run_until_complete(proxy_req.close())
         proxy_resp.close()
-        req.close()
+        self.loop.run_until_complete(req.close())
 
     @unittest.mock.patch('aiohttp.connector.ClientRequest')
     def test_https_connect_resp_start_error(self, ClientRequestMock):
@@ -1180,6 +1099,6 @@ class TestProxyConnector(unittest.TestCase):
         tr.pause_reading.assert_called_once_with()
         tr.get_extra_info.assert_called_with('socket', default=None)
 
-        proxy_req.close()
+        self.loop.run_until_complete(proxy_req.close())
         proxy_resp.close()
-        req.close()
+        self.loop.run_until_complete(req.close())
