@@ -172,22 +172,6 @@ class TestHttpMessage(unittest.TestCase):
 
         self.assertIn('SERVER', msg.headers)
 
-    def test_default_headers_useragent(self):
-        msg = protocol.Request(self.transport, 'GET', '/')
-        msg._add_default_headers()
-
-        self.assertNotIn('SERVER', msg.headers)
-        self.assertIn('USER-AGENT', msg.headers)
-
-    def test_default_headers_useragent_custom(self):
-        msg = protocol.Request(self.transport, 'GET', '/')
-        msg.add_headers(('user-agent', 'my custom agent'))
-        msg._add_default_headers()
-
-        headers = [r for r, _ in msg.headers.items()
-                   if r.lower() == 'user-agent']
-        self.assertEqual(len(headers), 1)
-
     def test_default_headers_chunked(self):
         msg = protocol.Response(self.transport, 200)
         msg._add_default_headers()
@@ -434,7 +418,9 @@ class TestHttpMessage(unittest.TestCase):
         msg.write(b'data')
         msg.write_eof()
 
-        content = b''.join([c[1][0] for c in list(write.mock_calls)])
+        chunks = [c[1][0] for c in list(write.mock_calls)]
+        self.assertTrue(all(chunks))
+        content = b''.join(chunks)
         self.assertEqual(
             self._COMPRESSED, content.split(b'\r\n\r\n', 1)[-1])
 
@@ -449,7 +435,9 @@ class TestHttpMessage(unittest.TestCase):
         msg.write(b'data')
         msg.write_eof()
 
-        content = b''.join([c[1][0] for c in list(write.mock_calls)])
+        chunks = [c[1][0] for c in list(write.mock_calls)]
+        self.assertTrue(all(chunks))
+        content = b''.join(chunks)
         self.assertEqual(
             b'2\r\nKI\r\n2\r\n,I\r\n2\r\n\x04\x00\r\n0\r\n\r\n',
             content.split(b'\r\n\r\n', 1)[-1])
@@ -466,7 +454,9 @@ class TestHttpMessage(unittest.TestCase):
         msg.write(b'data')
         msg.write_eof()
 
-        content = b''.join([c[1][0] for c in list(write.mock_calls)])
+        chunks = [c[1][0] for c in list(write.mock_calls)]
+        self.assertTrue(all(chunks))
+        content = b''.join(chunks)
         self.assertEqual(
             self._COMPRESSED, content.split(b'\r\n\r\n', 1)[-1])
 
